//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#include <vcl\clipbrd.hpp>
#pragma hdrstop

#include "MsgInWin.h"
#include "MsgWins.h"
#include "NoteWins.h"
#include "MailboxWins.h"
#include "AddressDlgs.h"
#include "PickBoxDlgs.h"
#include "FileMan.h"
#include "global.h"
#include "MixmastDlg.h"
#include "Mixmaster.h"
#include "PgpMsg.h"
#include "PGPpassEx.h"
#include "clipreply.h"
#include "pop3mail.h"
#include "savedlg.h"
#include "pgp.h" // for pgp defines only
#include "pgpmail.h"
#include "pgpfiles.h"
#include "helpdlgs.h"
#include "view.h"
//---------------------------------------------------------------------------
#pragma link "TextWins"
#pragma link "explbtn"
#pragma resource "*.dfm"
TMessageInWin *MessageInWin;
//---------------------------------------------------------------------------
bool __fastcall OpenMessageIn(AnsiString &MailBox, AnsiString &IndexStr)
{
   if (MailBox.IsEmpty() || IndexStr.IsEmpty()) return false;
   MessageInWin = new TMessageInWin(MailBox,IndexStr,AppMainForm);
   return true;
}
//---------------------------------------------------------------------------
__fastcall TMessageInWin::TMessageInWin(TComponent* Owner)
  : TTextWin(Owner)
{
}
//---------------------------------------------------------------------------
__fastcall TMessageInWin::TMessageInWin(AnsiString &MailBox, AnsiString &Indexstr,
   TComponent* Owner) : TTextWin(Owner)
{
   Type = tInMessage;
   Mailbox = MailBox;
   IndexStr = Indexstr;
   DataKey = UserMailboxes+PS+Mailbox+PS+IndexStr;
   Headers = new THeaderManager(Editor->Lines);
   HeadersBtnSet = false;
   WordWrapSet = false;
   Editor->Modified = false;
   SetEditMode(false);
   SetTrashBtnPos();
   SetMessage();
   Shifted = false;
}
//---------------------------------------------------------------------------
__fastcall TMessageInWin::~TMessageInWin()
{
   delete Headers;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetNoMixmaster()
{
   PGPcdecBtn->Visible = false;
   Bevel99->Visible = false;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetNoPgp()
{
   TrashBtn->Visible = true;
   TransferBtn->Visible = true;
   PGPcdecBtn->Visible = true;
   Bevel99->Visible = true;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetEditButtons(bool enable)
{
   long sel = Editor->SelLength;
   if (enable)
   {
      CutBtn->Enabled = sel && EditBtn->Down;
      CopyBtn->Enabled = sel;
      PasteBtn->Enabled = Clipboard()->HasFormat(CF_TEXT) && EditBtn->Down;
   }
   else
   {
      PasteBtn->Enabled = false;
      CopyBtn->Enabled = false;
      CutBtn->Enabled = false;
   }
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetTrashBtnPos()
{
   HelpBtn->Left = Width-(HelpBtn->Width+10);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetMessage()
{
   SetRedraw(Editor,false);
   TStrings *lines = Editor->Lines;
   Editor->OnSelectionChange = NULL;
   AnsiString cap=IndexStr+" ",fs=FieldSep,ic=InCode;
 //change caption to remove markers===
   for (int i=1; i<=cap.Length(); i++)
      if (cap[i]==fs[1] || cap[i]==ic[1]) cap[i]=' ';
   cap = cap.Trim();
   while (cap.Length() && cap[cap.Length()]!=' ') cap.SetLength(cap.Length()-1);
   Caption = Mailbox+": "+cap;
 //=====================================================
   FullHeaders = false;
   Editor->ReadOnly = true;
   Editor->WordWrap = false;
   // handle status change
   if (GetMsgStatus(DataKey)=='N') SetMsgStatus(DataKey,' ');
   LoadMessage(lines,FullHeaders);
   if (!HasPgp()) SetNoPgp();
   if (!HasMix()) SetNoMixmaster();
   else
   {
      FixPgpBeginEndLines();
      PGPcdecBtn->Enabled = lines->IndexOf(BeginPgpMessage)>=0; // || lines->IndexOf(BeginPgpSignedMessage)>=0;
      Editor->Modified = false;
      Editor->SelStart = 0;
   }
   HeadersBtn->Enabled = true;
   Editor->OnSelectionChange = EditorSelectionChange;
   SetRedraw(Editor,true);
}
//---------------------------------------------------------------------------
int eohm(TStrings *msg)
{
   int e=0; while (e<msg->Count && !msg->Strings[e].IsEmpty()) e++;
   return e;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::CreateReply(bool UseDialog)
{
   AnsiString TemplateName=DefaultMessage,inreplyto;
   if (UseDialog && ((TemplateName=GetMergeMessage("Reply using...",modeMessage))).IsEmpty()) return;
   if (GetMsgStatus(DataKey)==' ')
   {
      SetMsgStatus(DataKey,'R');
      inreplyto = DataKey;
   }
   TStringList *inmsg = new TStringList();
   GetAppData(inmsg,DataKey);//Editor->Lines->Text;
   // see if there is an email address selected
   AnsiString ToAddr = Editor->SelText;
   if (ToAddr.Pos("@") && ToAddr.Pos(".") && ToAddr.Pos("@") < ToAddr.Pos("."))
   {
      // just replace the 'From:' header--this will be converted
      // to the reply 'to:' address in ReplyToMessage function
      THeadMan *hm = new THeadMan(inmsg);
      int frhdr = hm->Delete("From: ");
      inmsg->Insert(frhdr>=0?frhdr:0,"From: "+ToAddr);
   }
   ReplyToMessage(inmsg,TemplateName,inreplyto);
   delete inmsg;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::HeadersBtnClick(TObject *Sender)
{
   if (EditBtn->Down)
   {
      HeadersBtn->Down = true;
      MyBeep(MB_ICONSTOP);
      return;
   }
   FullHeaders = HeadersBtn->Down;
   LoadMessage(Editor->Lines,FullHeaders);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::ReplyBtnClick(TObject *Sender)
{
   CreateReply(true);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::FormCloseQuery(TObject *Sender, bool &CanClose)
{
   if (Editor->Modified && !CheckSave())
   {
      CanClose = false;
      return;
   }
   if (MDIChildCount==0) SetParentButtons(false);
   ShowCurXY(false);
}
//---------------------------------------------------------------------------
bool __fastcall TMessageInWin::CheckSave()
{
   bool ok = true;
   if (Editor->Modified)
   {
      int reply = MsgBoxHwnd(Handle,"Save changes to message?",
         "Save Message",MB_YESNOCANCEL);
      switch(reply)
      {
         case IDCANCEL: ok = false; break;
         case IDYES: Save(); break;
      }
   }
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TMessageInWin::Save()
{
   AnsiString Cap = Caption;
   Caption = "Saving...";
   AppProcessMessages(); Sleep(500); Caption = Cap;

   TStringList *lines = new TStringList();
   GetUnWrappedText(lines);
   PutAppData(DataKey,lines);
   delete lines;

   HeadersBtn->Down = false;
   Editor->Modified = false;
// Editor->SelStart = 0;
   EditBtn->Down = false;
   Type = tMessage;
   SetMessage();

   AnsiString shomsg = Caption+" ... saved!";
   AppMainForm->Perform(WM_SHOWSTATUS,0,(LPARAM)shomsg.c_str());
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TMessageInWin::SaveAs()
{
   // not in use--don't save incoming msg to drafts
   bool ok=false;
   AnsiString Cap,path = GetSaveMessagePathEx(UserMessages+"\n"+UserNotes,Name());
   if (!path.IsEmpty())
   {
      Cap = Caption;
      Caption = "Saving...";
      AppProcessMessages(); Sleep(500);

      DataKey = path;
      TStringList *lines = new TStringList();
      GetUnWrappedText(lines);
      PutAppData(DataKey,lines);
      delete lines;

      Type = tMessage;
      SetMessage();

      Caption = GetParam(0,UserMessages+": "+GetParam(1,DataKey,PS.c_str()));
      ok = true;
   }
   return ok;
}
//---------------------------------------------------------------------------
bool __fastcall TMessageInWin::SaveAsNote()
{
   AnsiString Cap,path = GetSaveMessagePathEx(UserNotes,Name());
   if (!path.IsEmpty())
   {
      DataKey = path;
      Cap = Caption;
      Caption = "Saving...";
      AppProcessMessages(); Sleep(200);
      TStringList *lines = new TStringList();
      GetUnWrappedText(lines);
      PutAppData(DataKey,lines);
      Editor->Clear(); AppProcessMessages(); Sleep(200);
      path = GetParam(0,UserNotes+": "+GetParam(1,DataKey,PS.c_str()));
      NoteWin = new TNoteWin(Left,Top,Width,Height,path,lines,AppMainForm);
      Editor->Modified = false;
      Close();
      return true;
   }
   return false;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::OkBtnClick(TObject *Sender)
{
//   if (!Editor->Modified || (Editor->Modified && Save()))
   Close();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::WordWrapClick(TObject *Sender)
{
/*   TStringList *tmp = new TStringList();
   bool locked = LockWindowUpdate(Handle);
   if (Editor->WordWrap) Editor->WordWrap = false;
   tmp->Text = Editor->Lines->Text; Editor->Clear();
   Editor->WordWrap = WordWrap->Checked;
   Editor->Lines->Text = tmp->Text;
   Editor->SelStart = 0;
   delete tmp;
   SetEditRect();
//   for (int i=0; i<32; i++) AppProcessMessages();
   AppProcessMessages();
   if (locked) LockWindowUpdate(0);*/
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::EditBtnClick(TObject *Sender)
{
   SetRedraw(this,false);
   SetEditMode(EditBtn->Down);
   long sel = Editor->SelLength;
   if (EditBtn->Down)
   {
      CutBtn->Enabled = sel;
      CopyBtn->Enabled = sel;
      PasteBtn->Enabled = Clipboard()->HasFormat(CF_TEXT);
   }
   else
   {
      PasteBtn->Enabled = false;
      CopyBtn->Enabled = sel;
      CutBtn->Enabled = false;
   }
   SetRedraw(this,true);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::LoadMessage(TStrings *msg, bool fullheaders)
{
   TStringList *tmp = new TStringList();
   GetAppData(tmp,DataKey);
   if (!fullheaders)
   {
      AnsiString hdr;
      TStringList *hdrs = new TStringList();
      TStringList *minhdrs = new TStringList();
      minhdrs->Add("Date: ");
      minhdrs->Add("From: ");
      minhdrs->Add("Comments: ");
      minhdrs->Add("To: ");
      minhdrs->Add("Cc: ");
      minhdrs->Add("Subject: ");

      Headers->HereIs(tmp);
      while (!(hdr=Headers->PopHeader()).IsEmpty())
      {
         if (minhdrs->IndexOf(hdr.SubString(1,hdr.Pos(": ")+1))>=0)
            hdrs->Add(hdr);
      }
      tmp->Delete(0);
      tmp->Insert(0,hdrs->Text);
      delete minhdrs;
      delete hdrs;
   }
   Editor->OnSelectionChange = NULL;
   msg->Text = tmp->Text;
   Editor->OnSelectionChange = EditorSelectionChange;
   Editor->Modified = false;
   Editor->SelStart = 0;
   delete tmp;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SetEditMode(bool Editing)
{
   SetRedraw(Editor,false);
   if (Editing)
   {
      EditBtn->Down = false;
      HeadersBtn->Down = true;
      HeadersBtn->Click();
      EditBtn->Down = true;
      ReplyBtn->Enabled = false;
      TrashBtn->Enabled = false;
      TransferBtn->Enabled = false;
      PopupCopyItem->Enabled = false;
      PopupDeleteItem->Enabled = false;
      PopupTransferItem->Enabled = false;
      Editor->WordWrap = false;
      Editor->ReadOnly = false;
      Editor->Modified = false;
      HeadersBtn->Down = true;
      FullHeaders = true;
   }
   else
   {
      if (CheckSave())
      {
         if (!HeadersBtnSet) // HeadersBtnSet is always false--dropped idea tracking
         {
            HeadersBtn->Down = false;
            HeadersBtn->Click();
         }
         ReplyBtn->Enabled = true;
         TrashBtn->Enabled = true;
         TransferBtn->Enabled = true;
         PopupCopyItem->Enabled = true;
         PopupDeleteItem->Enabled = true;
         PopupTransferItem->Enabled = true;
         Editor->WordWrap = false;//WordWrapSet;
         Editor->ReadOnly = true;
         PasteBtn->Enabled = false;
         CutBtn->Enabled = false;
      }
      else
         EditBtn->Down = true;
   }
   SetRedraw(Editor,true);
}
//---------------------------------------------------------------------------
AnsiString __fastcall PgpErrorStr(int errcode)
{
   AnsiString msg;
   switch (errcode)
   {
      case PGP_OK: break; // success
      case PGP_SIGOK: break; // success, and signature has been verified
      case PGP_ERR:     msg = "The message can't be read!"; break;
      case PGP_PASS:    msg = "Bad passphrase"; break;
      case PGP_NOMSG:   msg = "Message is not an OpenPGP message"; break;
      case PGP_SIGNKEY: msg = "Can't verify signature"; break;
      case PGP_SIGBAD:  msg = "Bad signature"; break;
      case PGP_NODATA:  msg = "Message does not contain user data."; break;
   }
   return msg;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::FixPgpBeginEndLines()
{
   int idx;
   TStrings *lines=Editor->Lines;
   while((idx=lines->IndexOf("- "+BeginPgpMessage))>=0)
      lines->Strings[idx] = BeginPgpMessage;
   while((idx=lines->IndexOf("- "+EndPgpMessage))>=0)
      lines->Strings[idx] = EndPgpMessage;
   while((idx=lines->IndexOf("- "+BeginPgpSignedMessage))>=0)
      lines->Strings[idx] = BeginPgpSignedMessage;
   while((idx=lines->IndexOf("- "+BeginPgpSignature))>=0)
      lines->Strings[idx] = BeginPgpSignature;
   while((idx=lines->IndexOf("- "+EndPgpSignature))>=0)
      lines->Strings[idx] = EndPgpSignature;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PGPcdecBtnClick(TObject *Sender)
{
   int i,errcode=0,start,end=-1;
   AnsiString passphrase;
   bool ok=true,isconv=true,triednyms=false,clearsigned=false;//cnv;
   TStringList *pgpblock = new TStringList();
   TStringList *fullmsg = new TStringList();
   GetAppData(fullmsg,DataKey);
   fullmsg->Add(""); // so inserts can be made without checks

   FixPgpBeginEndLines();
 /*start = fullmsg->IndexOf(BeginPgpSignedMessage);
   if (start>=0)
   {
      end = fullmsg->IndexOf(EndPgpSignature);
      if (start<0 || end<=start)
      {
         MsgBoxHwnd(Handle,
           AnsiString("Unable to verify this message!\n\n")+
           AnsiString("It has some appearances of being a PGP\n")+
           AnsiString("signed message, but I have found a problem\n")+
           AnsiString("in the beginning and/or ending PGP markers."),
          "PGP Error",MB_OK
         );
         delete fullmsg;
         delete pgpblock;
         return;
      }
      clearsigned = true;
   }*/
   if (!clearsigned)
   {
      start = fullmsg->IndexOf(BeginPgpMessage);
      if (start>=0) end = fullmsg->IndexOf(EndPgpMessage);
      if (start<0 || end<=start)
      {
         MsgBoxHwnd(Handle,
           AnsiString("Unable to decrypt message!\n\n")+
           AnsiString("This message has some appearances of being\n")+
           AnsiString("PGP encrypted, but I have found a problem in\n")+
           AnsiString("the beginning and/or ending PGP markers."),
          "PGP Error",MB_OK
         );
         delete fullmsg;
         delete pgpblock;
         return;
      }
   }
   for (i=start; i<=end; i++)
   {
      pgpblock->Add(fullmsg->Strings[start]);
      fullmsg->Delete(start);
   }

   if (clearsigned)
   {
      PgpVerifyMessage(pgpblock->Text.c_str());
      return;
   }

decrypt:
   if (HasPgp())
   { // try a secret key decryption
      isconv = false;
      TStringList *msg = new TStringList();
      msg->SetText(PgpDecryptVerifyMessage(pgpblock->Text.c_str(),Shifted));
      if (msg->Text==IsConventionalEncr+"\r\n") isconv = true; else
      if (msg->Text.IsEmpty()) ok = false; else
      pgpblock->Text = msg->Text;
      delete msg;
   }

   if (isconv)
   {
      if (HasNyms() && NYMAUTODECRYPT && HasActualNyms() && !triednyms)
      {
         triednyms = true;
         TStringList *tmp = new TStringList();
         tmp->SetText(DecryptNymMessage(AppMainForm,pgpblock->Text.c_str(),errcode));
         if (tmp->Count)
         {
            pgpblock->Text = tmp->Text;
            goto decrypt;
         }
      }
      // try conventional decryption
      ok = PGPGETDecryptionPassphrase(this,passphrase);
      if (ok && (!(PgpDecryptConventional(pgpblock,passphrase,errcode)==mrOk) || errcode))
      {
         MsgBoxHwnd(Handle,"Unable to decrypt! "+PgpErrorStr(errcode),
            "PGP Error",MB_OK);
         ok = false;
      }
   }

   if (ok) // something worked !
   {
      for (i=0; i<pgpblock->Count; i++)
         fullmsg->Insert(start+i,pgpblock->Strings[i]);
      fullmsg->Delete(fullmsg->Count-1); // dump the trailing null line
      Editor->Clear(); AppProcessMessages(); //Sleep(100);
      SetRedraw(this,0);
      Editor->Lines->Text = fullmsg->Text;
      if (GetAppIni(SaveMsgClear,"")=="1")
      {
         Save();
//       GetAppData(Editor->Lines,DataKey);
         SetMessage();
      }
      else
      {
         HeadersBtn->Enabled = false;
         EditBtn->Enabled = false;
      }

      PGPcdecBtn->Enabled = Editor->Lines->IndexOf(BeginPgpMessage);
      Editor->Modified = false;
      Editor->SelStart = 0;
      SetRedraw(this,1);
   }
   // tidy up
   delete pgpblock;
   delete fullmsg;
}
/*
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PGPcdecBtnClick(TObject *Sender)
{
   bool ok=true;
   bool isconv=false;
   int errcode=0;
   AnsiString passphrase;
   TStringList *fullmsg = new TStringList();
   TStringList *pgpblock = new TStringList();
   GetAppData(fullmsg,DataKey);
   fullmsg->Add(""); // so inserts can be made without checks
   int i,end,start = fullmsg->IndexOf("-----BEGIN PGP MESSAGE-----");
   if (start) end = fullmsg->IndexOf("-----END PGP MESSAGE-----");
   if (start>=0 && end>start)
   {
      for (i=start; i<=end; i++)
      {
         pgpblock->Add(fullmsg->Strings[start]);
         fullmsg->Delete(start);
      }
      if (!HasPgp() || PgpIsConventional(pgpblock->Text.c_str()))
      {
         ok = PGPgetDecryptionPassphrase(passphrase,AppMainForm);
         if (ok && (!(PgpDecryptConventional(pgpblock,passphrase,errcode)==mrOk) || errcode))
         {
            MsgBoxHwnd(Handle,"Unable to decrypt! "+PgpErrorStr(errcode),
               "PGP Error",MB_OK);
            ok = false;
         }
      }
      else // has PGP
      {
         TStringList *msg = new TStringList();
         msg->SetText(PgpDecryptVerifyMessage(pgpblock->Text.c_str()));
         if (msg->Text.IsEmpty()) ok = false;
         else pgpblock->Text = msg->Text;
         delete msg;
      }
      // if either operation was successful
      if (ok)
      {
         //pgpblock->Insert(0,"-----BEGIN DECRYPTED PGP MESSAGE-----");
         //pgpblock->Add("-----END DECRYPTED PGP MESSAGE-----");
         for (i=0; i<pgpblock->Count; i++)
            fullmsg->Insert(start+i,pgpblock->Strings[i]);
         fullmsg->Delete(fullmsg->Count-1); // dump the trailing null line

         Editor->Clear(); AppProcessMessages(); Sleep(200);
         SetRedraw(this,false);
         Editor->Lines->Text = fullmsg->Text;
         if (GetAppIni(SaveMsgClear,"")=="1")
         {
            Save();
            GetAppData(Editor->Lines,DataKey);
            SetMessage();
         }
         else
         {
            HeadersBtn->Enabled = false;
            EditBtn->Enabled = false;
         }
         //Editor->Perform(EM_LINESCROLL,0,start);
         //Editor->SelStart = Editor->Perform(EM_LINEINDEX,start,NULL);
         //EditBtn->Down = true;
         PGPcdecBtn->Enabled = false;
         Editor->Modified = false;
         Editor->SelStart = 0;
         SetRedraw(this,true);
      }
   }
   else
   {
      MsgBoxHwnd(Handle,
        AnsiString("Unable to decrypt message!\n\n")+
        AnsiString("This message has some appearances of being\n")+
        AnsiString("PGP encrypted, but I have found a problem in\n")+
        AnsiString("the beginning and/or ending PGP markers."),
       "PGP Error",MB_OK
      );
   }
   delete pgpblock;
   delete fullmsg;
}
*/
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SaveBtnClick(TObject *Sender)
{
  TPoint cpoint,poppoint={ -2,27 };
  cpoint = OkBtn->ClientToScreen(poppoint);
  PopupMenu1->Popup(cpoint.x,cpoint.y);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupMenu1Popup(TObject *Sender)
{
   PopupSaveItem->Enabled = Editor->Modified;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupSaveItemClick(TObject *Sender)
{
   Save();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupSaveasItemClick(TObject *Sender)
{
   SaveAs();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::TransferMessageTo(AnsiString &Boxname)
{
   if (Boxname.IsEmpty()) return;
   TStringList *fullmsg = new TStringList();
   GetAppData(fullmsg,DataKey);
   if (IndexStr.Length()>=24 && IndexStr[24]=='N') IndexStr[24] = ' ';

   AnsiString NewDataKey=UserMailboxes+PS+Boxname+PS+IndexStr;
   PutAppData(NewDataKey,fullmsg);
   PutAppData(DataKey,NULL);
   DataKey = NewDataKey;
   Mailbox = Boxname;
   delete fullmsg;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::CopyMessageTo(AnsiString &Boxname)
{
   if (Boxname.IsEmpty()) return;

   TStringList *fullmsg = new TStringList();
   GetAppData(fullmsg,DataKey);
   DataKey=UserMailboxes+PS+Boxname+PS+IndexStr;
   PutAppData(DataKey,fullmsg);
   Editor->Clear(); AppProcessMessages(); Sleep(200);
   SetRedraw(this,false);
   GetAppData(Editor->Lines,DataKey);
   Mailbox = Boxname;
   delete fullmsg;
   SetMessage();
   SetRedraw(this,true);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupTransferItemClick(TObject *Sender)
{
   AnsiString Oldbox=Mailbox,Newbox = GetTransferBoxname("");
   if (Newbox.IsEmpty()) return;
   TransferMessageTo(Newbox);
   Close();
   // check to see if boxes are on desktop
   TMailboxWin *mbw=FindOpenMailbox(Oldbox);
   if (mbw!=NULL) mbw->RefreshIndex();
   mbw=FindOpenMailbox(Newbox);
   if (mbw!=NULL) mbw->RefreshIndex();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupCopyItemClick(TObject *Sender)
{
   AnsiString Oldbox=Mailbox,Newbox=GetTransferBoxname("");
   if (Newbox.IsEmpty()) return;

   CopyMessageTo(Newbox);
   // check to see if boxes are on desktop
   TMailboxWin *mbw=FindOpenMailbox(Newbox);
   if (mbw!=NULL) mbw->RefreshIndex();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupDeleteItemClick(TObject *Sender)
{
   if (MsgBox("Delete this message from the system?","Confirm",MB_OKCANCEL)==IDOK)
   {
      Caption=" Deleting..."; Sleep(100);
      AppProcessMessages();
      PutAppData(DataKey,NULL);

      TMailboxWin *mbw=FindOpenMailbox(Mailbox);
      if (mbw!=NULL) mbw->RefreshIndex();

      Editor->Modified = false;
      Close();
   }
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::OkBtnEnter(TObject *Sender)
{
   if (!SaveBtn->HasMouse()) SaveBtn->Perform(CM_MOUSEENTER,0,NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::OkBtnExit(TObject *Sender)
{
   if (SaveBtn->HasMouse()) SaveBtn->Perform(CM_MOUSELEAVE,0,NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SaveBtnEnter(TObject *Sender)
{
   if (!OkBtn->HasMouse()) OkBtn->Perform(CM_MOUSEENTER,0,NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SaveBtnExit(TObject *Sender)
{
   if (OkBtn->HasMouse()) OkBtn->Perform(CM_MOUSELEAVE,0,NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SaveMessageBtnClick(TObject *Sender)
{
   if (Save())
   {
      //SaveMessageBtn->Enabled = false;
      EditBtn->Down = false;
      EditBtn->Click();
   }
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupMenuClick(TObject *Sender)
{
   TTextWin::PopupMenuClick(Sender);
   TMenuItem *item = dynamic_cast<TMenuItem*>(Sender);
   if (item && item->Caption=="&Word wrap this message")
   {
      bool locked = LockWindowUpdate(Handle);
      PopWordWrapItem->Checked = !PopWordWrapItem->Checked;
      if (Editor->WordWrap != PopWordWrapItem->Checked)
         Editor->WordWrap = PopWordWrapItem->Checked;
      if (locked) LockWindowUpdate(0);
   }
   PopPasteItem->Enabled = PopPasteItem->Enabled && EditBtn->Down;
   PopCutItem->Enabled = PopCutItem->Enabled && EditBtn->Down;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::SaveasNote1Click(TObject *Sender)
{
   SaveAsNote();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::TransferItemOnClick(TObject *Sender)
{
   TMenuItem *item = dynamic_cast<TMenuItem*>(Sender);
   if (!item || item->Caption.IsEmpty()) return;
   AnsiString Oldbox=Mailbox,Newbox=item->Caption;
   if (Newbox=="&New...") Newbox = CreateMailbox();
   if (Newbox.IsEmpty()) return;
   Caption = "Transfering...";
   AppProcessMessages(); Sleep(100);
   TransferMessageTo(Newbox);
   Close();
   // check to see if boxes are on desktop
   TMailboxWin *mbw=FindOpenMailbox(Oldbox);
   if (mbw!=NULL) mbw->RefreshIndex();
   mbw=FindOpenMailbox(Newbox);
   if (mbw!=NULL) mbw->RefreshIndex();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::PopupMenu2Popup(TObject *Sender)
{
   int i;
   AnsiString s;
   TMenuItem *newitem;
   TStringList *boxes = new TStringList(); boxes->Sorted = true;
   GetAppData(boxes,MailboxBoxmenu);
   int precount = boxes->Count;
   i=0; while (i<boxes->Count)
   {
      s = boxes->Strings[i];
      if (s.IsEmpty() || s==UserMailboxes)
         boxes->Delete(i); else i++;
   }
   // test box list against list of all current boxes, delete if not there.
   TStringList *allboxes = new TStringList();
   GetAppData(allboxes,UserMailboxes);
   i=0; while (i<boxes->Count)
   {
      if (allboxes->IndexOf(boxes->Strings[i])<0)
         boxes->Delete(i); else i++;
   }
   if (boxes->Count!=precount) PutAppData(MailboxBoxmenu,boxes);
   delete allboxes;

   while (PopupMenu2->Items->Count>0) PopupMenu2->Items->Delete(0);

   newitem = new TMenuItem(this);
   newitem->OnClick = TransferItemOnClick;
   newitem->Caption = "&New...";
   PopupMenu2->Items->Add(newitem);

   newitem = new TMenuItem(this);
   newitem->OnClick = NULL;
   newitem->Caption = "-";
   PopupMenu2->Items->Add(newitem);

   for (i=0; i<boxes->Count; i++)
   {
      newitem = new TMenuItem(this);
      newitem->OnClick = TransferItemOnClick;
      newitem->Caption = boxes->Strings[i];
      PopupMenu2->Items->Add(newitem);
   }
   delete boxes;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::TrashBtnClick(TObject *Sender)
{
   PopupDeleteItemClick(NULL);
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::CopyAddresstoAddressBook1Click(TObject *Sender)
{
   AnsiString &address = Editor->SelText;
   if (!address.Pos("@")) address = Headers->GetDat("From: ",Editor->Lines);
   if (!address.Pos("@")) return;

   if (MsgBoxHwnd(Handle,
      s("Copy this address to your Address Book?\n\n")+address,"Confirm",
      MB_OKCANCEL)==IDCANCEL) return;

   TStringList *addrs = new TStringList(); addrs->Sorted = true;
   GetAppData(addrs,UserAddresses+PS+General);
   addrs->Add(address);
   PutAppData(UserAddresses+PS+General,addrs);
   delete addrs;
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::FormResize(TObject *Sender)
{
   TTextWin::FormResize(Sender);
   SetTrashBtnPos();
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::HelpBtnClick(TObject *Sender)
{
   AnsiString topics = "MessageIn";
   if (HasPgp()) topics += "\nMessageInPGP\nPgpPassphrase";
   if (HasNyms()) topics += "\nMessageInNym";
   OpenHelp(topics,"Incoming Mail","");
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::FormKeyDown(TObject *Sender, WORD &Key,
	TShiftState Shift)
{
   Shifted = true;	
}
//---------------------------------------------------------------------------
void __fastcall TMessageInWin::FormKeyUp(TObject *Sender, WORD &Key,
	TShiftState Shift)
{
   Shifted = false;
}
//---------------------------------------------------------------------------
