//---------------------------------------------------------------------------
#include <vcl\vcl.h>
#pragma hdrstop

#include "PGPbuffers.h"
#include "PGPfiles.h"
#include "PGPmail.h"
#include "mixmaster.h" //BUFFER functions
#include "fileman.h"
#include "view.h"
//---------------------------------------------------------------------------
// TPGPbuffer
//---------------------------------------------------------------------------
__fastcall TPGPbuffer::TPGPbuffer()
   : TBUFFER()
{
}
//---------------------------------------------------------------------------
__fastcall TPGPbuffer::TPGPbuffer(AnsiString keyringfile)
   : TBUFFER()
{
   loadfromfile(keyringfile);
}
//---------------------------------------------------------------------------
__fastcall TPGPbuffer::TPGPbuffer(TStrings *asciiblock)
   : TBUFFER()
{
   import(asciiblock);
   rewind();
}
//---------------------------------------------------------------------------
__fastcall TPGPbuffer::TPGPbuffer(TPGPbuffer *buf)
   : TBUFFER()
{
   buf->rewind();
   copyfrom(buf);
   rewind();
}
//---------------------------------------------------------------------------
__fastcall TPGPbuffer::~TPGPbuffer()
{
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::loadfromfile(AnsiString keyringfile)
{
   clear();
   filename = keyringfile;
   if (FileExists(filename))
   {
      BUFFER *buf=buf_new(),*armored=buf_new(),*keybuf=buf_new(),*asckeys=buf_new();
      buf_readfile(armored,filename.c_str());
      if (pgpispacket(armored)) buf_move(keybuf,armored);
      else
      {
         while (GetNextPGPArmored(asckeys,armored))
         {
            pgpdearmor(asckeys,buf);
            buf_cat(keybuf,buf);
         }
      }
      Write(keybuf->data,keybuf->length);
      buf_free(asckeys);
      buf_free(armored);
      buf_free(keybuf);
      buf_free(buf);
   }
   rewind();
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::parsekeycert()
{
   bool ok=true;
   if (fparsed) return ok;

   if (IsPublic()) GetIdText8(); else
   if (IsSecret())
   {
      BUFFER *packet = buf_new();
      buf_append(packet,data,length);
      ok=ParseSecretKeyPacket(packet,&fidtext8,&fversion,&funixtime,&fvalid,&falgo,
         &fmodbits,&fexpbits,NULL,NULL,&falgocip,NULL,NULL,&fdexbits,
         NULL,&ffacpbits,NULL,&ffacqbits,NULL,&finvbits,&fchecksum);
      buf_free(packet);
   }
   fparsed = true;
   return ok;
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::onlyusefulkeys() // removes disables/revokes/expires
{
   int cnt=0;
   TPGPbuffer *key = new TPGPbuffer();
   TPGPbuffer *keys = new TPGPbuffer();
   rewind(); while (key->copykeyfrom(this,REWIND))
   {
      if (key->enabled && !key->expired && !key->revoked)
      {
         keys->copyfrom(key);
         cnt++;
      }
      key->clear();
   }
   clear(); copyfrom(keys); rewind();
   delete keys;
   delete key;
   return cnt;
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::clear()
{
   TBUFFER::clear();
   ffacpbits = ffacqbits = finvbits = fchecksum = fsubbits = 0;
   funixtime = fvalid = fmodbits = fexpbits = falgocip = 0;
   fversion = falgo = fsubalgo = 0;
   fidtext8 = "";
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::IsPublic()
{
   return PublicKeyPacket(CTB);
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::IsSecret()
{
   return SecretKeyPacket(CTB);
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::export()
{
   AnsiString exp;
   TStringList *key = new TStringList();
   export(key);
   exp = key->Text;
   delete key;
   return exp;
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::export(TStrings *asciiblock)
{
   BUFFER *asc = buf_new();
   TPGPbuffer *pkt = new TPGPbuffer();
   TPGPbuffer *keys = new TPGPbuffer();
   rewind(); while(pkt->copypacketfrom(this,REWIND))
   {
      if (!TrustPacket(pkt->CTB)) keys->copyfrom(pkt);
      pkt->clear();
   }
   buf_append(asc,keys->data,keys->length);
   pgparmor(asc,2);
   asciiblock->SetText((char*)asc->data);
   delete keys;
   delete pkt;
   buf_free(asc);
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::import(TStrings *asciiblock)
{
   if (asciiblock->Count)
   {
      BUFFER *buf=buf_new(),*armored=buf_new(),*keybuf=buf_new(),*asckeys=buf_new();
      TStringList *keys = new TStringList();
      keys->Text = asciiblock->Text;
      while (keys->Count && !IsPGPArmoredHead(keys->Strings[0]))
         keys->Delete(0);
      if (keys->Count)
      {
         buf_sets(armored,asciiblock->Text.c_str());
         if (pgpispacket(armored)) buf_move(keybuf,armored);
         else
         {
            while (GetNextPGPArmored(asckeys,armored))
            {
               pgpdearmor(asckeys,buf);
               buf_cat(keybuf,buf);
            }
         }
         Write(keybuf->data,keybuf->length);
      }
      buf_free(asckeys);
      buf_free(armored);
      buf_free(keybuf);
      buf_free(buf);
      delete keys;
   }
   rewind();
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::FirstUID()
{
   byte CTB;
   AnsiString uid;
   char buffer[256];
   long pktlen,Ptr=ptr;
   TPGPbuffer *buf = new TPGPbuffer();
   rewind(); while (buf->copypacketfrom(this))
   {
      pktlen = buf->GetPacketInfo(&CTB);
      if (UserIdPacket(CTB))
      {
         buf->Read(buffer,pktlen); buffer[pktlen]=0L;
         uid = buffer;
         break;
      }
      buf->clear();
   }
   ptr = Ptr;
   delete buf;
   return uid;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::keyexists(TPGPbuffer *key)
{
   long Ptr=ptr;
   bool exists = false;
   key->rewind(); rewind();
   TPGPbuffer *buf = new TPGPbuffer();
   while (buf->copykeyfrom(this))
   {
      if (buf->idtext8==key->idtext8)
      {
         exists = true;
         break;
      }
      buf->clear();
   }
   key->rewind();
   delete buf;
   ptr = Ptr;
   return exists;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::getidtext8(AnsiString UserId)
{
   rewind();
   AnsiString idtext8;
   TPGPbuffer *key = new TPGPbuffer();
   while (key->copykeyfrom(this))
   {
      if (key->getuserids().LowerCase().Pos(UserId)>0)
      {
         idtext8 = key->idtext8;
         break;
      }
      key->clear();
   }
   return idtext8;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::get(AnsiString IdText8, TPGPbuffer *key)
{
   long Ptr = ptr;
   bool got = false;
   AnsiString idtext8 = IdText8.Pos("0x")==1?IdText8.SubString(3,8):IdText8;
   rewind(); key->clear();
   while (key->copykeyfrom(this))
   {
      if (key->idtext8==idtext8)
      {
         key->rewind();
         got = true;
         break;
      }
      key->clear();
   }
   ptr = Ptr;
   return got;
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::getuserids(TStrings *ids)
{
   byte CTB;
   rewind();
   ids->Clear();
   AnsiString uid;
   char buffer[256];
   long pktlen,Ptr=ptr;
   TPGPbuffer *buf = new TPGPbuffer();
   rewind(); while (buf->copypacketfrom(this))
   {
      pktlen = buf->GetPacketInfo(&CTB);
      if (UserIdPacket(CTB))
      {
         buf->Read(buffer,pktlen); buffer[pktlen]=0L;
         ids->Add(buffer);
      }
      buf->clear();
   }
   delete buf;
   ptr = Ptr;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::hasuserid(AnsiString &uid)
{
   return getuserids().LowerCase().Pos(uid)>0;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::getuserids()
{
   AnsiString idsstr;
   TStringList *ids = new TStringList();
   getuserids(ids);
   idsstr = ids->Text;
   delete ids;
   return idsstr;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::getdescrips()
{
   TStringList *lines = new TStringList();
   getdescrips(lines);
   AnsiString desc = lines->Text;
   delete lines;
   return desc;
}
//---------------------------------------------------------------------------
int __fastcall longest(TStrings *lines)
{
   int x,len=0;
   for (int i=0; i<lines->Count; i++)
      if ((x=lines->Strings[i].Trim().Length())>len) len = x;
   return len;
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::getdescrips(TStrings *lines)
{
   rewind();
   TPGPbuffer *key = new TPGPbuffer();
   while (key->copykeyfrom(this,REWIND))
   {
      //key->parsekeycert();
      lines->Add("0x"+key->idtext8+"  "+key->firstuid.Trim());
      key->clear();
   }
   rewind();

   int i=0,hilen=longest(lines);
   while (key->copykeyfrom(this,REWIND))
   {
      AnsiString space,version,ispair,enabled,algo,bits,type,line;
      line = lines->Strings[i];
      //if (line.Length()<hilen+2)
      line += space.StringOfChar(32,(hilen)-line.Length());

      key->parsekeycert();
      version = s(key->version);
      enabled = key->enabled ? "enabled":"disabled";
      algo = key->algostr+space.StringOfChar(32,6-key->algostr.Length());
      bits = key->bitsstr; if (bits.Length()==3) bits = " "+bits;
      bits = bits+space.StringOfChar(32,9-bits.Length());
      type = ispublic ? "pub" : "sec";
      //ispair = key->ispair ? "pair  ":"pub   ";
      lines->Strings[i] = line+"  "+version+"  "+type+"  "+algo+"  "+bits+"  "+enabled;
      key->clear();

      i++;
   }
   delete key;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::enable(bool Enable)
{
   bool gotit=false;
   TPGPbuffer *pkt = new TPGPbuffer();
   TPGPbuffer *key = new TPGPbuffer(this);
   key->rewind();
   clear();
   pkt->copypacketfrom(key,REWIND); // get key cert
   copyfrom(pkt);                   // copy to self
   pkt->clear();
   pkt->copypacketfrom(key,REWIND); // copy next--trust
   if (TrustPacket(pkt->CTB))
   {  // if it is a trust packet set enable to desired
      byte val = pkt->data[2];
      if ( Enable &&   (val&32)==32)  pkt->data[2] = (byte)(val^32); else
      if (!Enable && !((val&32)==32)) pkt->data[2] = (byte)(val^32);
   }
   copyfrom(pkt); pkt->clear();
   while (pkt->copypacketfrom(key,REWIND))
   {
      copyfrom(pkt);
      pkt->clear();
   }
   delete key;
   delete pkt;
   return gotit;
}
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::disable()
{
   TPGPbuffer *pkt = new TPGPbuffer();
   TPGPbuffer *key = new TPGPbuffer(this);
   key->rewind();
   clear();
   pkt->copypacketfrom(key,REWIND);
   copyfrom(pkt);
   pkt->clear();
   pkt->copypacketfrom(key,REWIND);
   if (TrustPacket(pkt->CTB))
      pkt->data[2] = (byte)(pkt->data[2]^32);
   copyfrom(pkt); pkt->clear();
   while (pkt->copypacketfrom(key,REWIND))
   {
      copyfrom(pkt);
      pkt->clear();
   }
   delete key;
   delete pkt;
}
//---------------------------------------------------------------------------
/*bool __fastcall TPGPbuffer::revoke(bool Revoke)
{
   bool gotit=false;
   TPGPbuffer *pkt = new TPGPbuffer();
   TPGPbuffer *key = new TPGPbuffer(this);
   key->rewind();
   clear();
   pkt->copypacketfrom(key,REWIND); // get key cert
   copyfrom(pkt);                   // copy to self
   pkt->clear();
   pkt->copypacketfrom(key,REWIND); // copy next--trust
   if (TrustPacket(pkt->CTB))
   {  // if it is a trust packet set enable to desired
      byte val = pkt->data[2];
      if ( Revoke &&   (val&16)==16)  pkt->data[2] = (byte)(val^16); else
      if (!Revoke && !((val&16)==16)) pkt->data[2] = (byte)(val^16);
   }
   copyfrom(pkt); pkt->clear();
   while (pkt->copypacketfrom(key,REWIND))
   {
      copyfrom(pkt);
      pkt->clear();
   }
   delete key;
   delete pkt;
   return gotit;
}*/
//---------------------------------------------------------------------------
void __fastcall TPGPbuffer::revoke()
{
   TPGPbuffer *pkt = new TPGPbuffer();
   TPGPbuffer *key = new TPGPbuffer(this);
   key->rewind();
   clear();
   pkt->copypacketfrom(key,REWIND); // get key cert
   copyfrom(pkt);                   // copy to self
   pkt->clear();
   pkt->copypacketfrom(key,REWIND); // copy next--trust
   if (TrustPacket(pkt->CTB))
      pkt->data[2] = (byte)(pkt->data[2]^16);
   copyfrom(pkt); pkt->clear();
   while (pkt->copypacketfrom(key,REWIND))
   {
      copyfrom(pkt);
      pkt->clear();
   }
   delete key;
   delete pkt;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::save()
{
   if (filename.IsEmpty()) return false;
   SaveToFile(filename);
   return true;
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::Count()
{
   long c=0,Ptr=ptr;
   rewind();
   TPGPbuffer *key = new TPGPbuffer();
   while (key->copykeyfrom(this))
   {
      key->clear();
      c++;
   }
   delete key;
   ptr = Ptr;
   return c;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::IsPair()
{
   rewind();
   int pub=0,sec=0;
   TPGPbuffer *key = new TPGPbuffer();
   for (int i=0; i<2; i++)
   {
      if (key->copykeyfrom(this))
      {
         if (PublicKeyPacket(key->CTB)) pub++; else
         if (SecretKeyPacket(key->CTB)) sec++;
      }
      key->clear();
   }
   return pub>0 && sec>0;
}
//---------------------------------------------------------------------------
/*if (ss(trust,0,1)=="1") str+= "1-------  BUCKSTOP: Ultimately trusted keyring owner |";
else                      str+= "0-------  Public key certificate |";
if (ss(trust,2,1)=="1")   str+= "--1-----  Disabled user's request |";
else                      str+= "--0-----  Enabled |";
if (ss(trust,3,1)=="1")   str+= "---1----  Revoked at issuer's request |";
else                      str+= "---0----  Not revoked |";*/
bool __fastcall TPGPbuffer::Enabled()
{
   bool enabled=true;
   TPGPbuffer *key = this;
   TPGPbuffer *pkt = new TPGPbuffer();

   key->rewind();
   pkt->copypacketfrom(key,REWIND); pkt->clear();
   pkt->copypacketfrom(key,REWIND);
   if (TrustPacket(pkt->CTB))
      enabled = !((pkt->data[2]&32)==32);
   delete pkt;
   return enabled;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::Revoked()
{
   bool revoked=false;
   TPGPbuffer *key = this;
   TPGPbuffer *pkt = new TPGPbuffer();

   key->rewind();
   pkt->copypacketfrom(key,REWIND); pkt->clear(); // key certificate
   pkt->copypacketfrom(key,REWIND);               // trust packet
   if (TrustPacket(pkt->CTB))
      revoked = (pkt->data[2]&16)==16;
   delete pkt;
   return revoked;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::Expired()
{
   byte CTB;
   bool expired=false;
   TPGPbuffer *key = this;
   TPGPbuffer *pkt = new TPGPbuffer();
   key->rewind(); pkt->GetPacketInfo(&CTB);
   if (PublicKeyPacket(CTB))
   {
      byte keyver = pkt->ReadByte();
      long keytime = pkt->ReadLong();
      long keyval = keyver<4 ? pkt->ReadInt():0;
      if (keyval>0)
      {
         keyval = keytime+(keyval*86400);
         if (keyval<time(NULL)) expired = true;
      }
   }
   delete pkt;
   return expired;
}
//---------------------------------------------------------------------------
byte __fastcall TPGPbuffer::GetCTB()
{
   if (length) return *((byte*)&data[0]);
   return 0;
}
//---------------------------------------------------------------------------
byte __fastcall TPGPbuffer::ReadByte()
{
   byte b;
   Read(&b,1);
   return b;
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::ReadInt()
{
   int i;
   byte b[2];
   Read(b,2);
   i  = b[0] << 8;
   i += b[1];
   return i;
}
//---------------------------------------------------------------------------
long __fastcall TPGPbuffer::ReadLong()
{
   long l;
   byte b[4];
   Read(b,4);
   l  = b[0] << 24;
   l += b[1] << 16;
   l += b[2] << 8;
   l += b[3];
   return (l);
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::WriteByte(byte b)
{
   Write(&b,1);
   return 0;
}
//---------------------------------------------------------------------------
int  __fastcall TPGPbuffer::WriteInt(int i)
{
   WriteByte((byte)((i>>8)&255));
   WriteByte((byte)(i&255));
   return 0;
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::WriteLong(long l)
{
   WriteByte((byte)((l>>24)&255));
   WriteByte((byte)((l>>16)&255));
   WriteByte((byte)((l>>8)&255));
   WriteByte((byte)(l&255));
   return 0;
}
//---------------------------------------------------------------------------
byte __fastcall lolf(byte CTB)
{
   byte len=1;
   if ((CTB&1)==1) len = 2; else
   if ((CTB&2)==2) len = 4; else
   if ((CTB&3)==3) len = 0;
   return len;
}
//---------------------------------------------------------------------------
int  __fastcall TPGPbuffer::GetPacketSize()
{
   if (ptr>=length-1) return 0;

   long lol,len=0,l=0;
   byte CTB,tmp,*b=(byte*)&l;

   Read(&CTB,1); lol=lolf(CTB);
   Read(&l,lol);
   ptr -= lol+1;

   switch (lol)
   {
      case 1: len = *b; break;
      case 2: tmp=b[0]; b[0]=b[1]; b[1]=tmp; len = *(UINT*)&l; break;
      case 4: len = l; break;
   }
   if (len) len += lol+1;
   return len;
}
//---------------------------------------------------------------------------
int  __fastcall TPGPbuffer::GetPacketInfo(byte *CTB, byte *lol)
{
   int len,l=0;
   byte tmp,Lol,*b=(byte*)&l;

   rewind();
   Read(CTB,1); Lol=lolf(*CTB);
   Read(&l,Lol);
   switch (Lol)
   {
      case 1: len = *(byte*)&l; break;
      case 2: tmp=b[0]; b[0]=b[1]; b[1]=tmp; len = *(UINT*)&l; break;
      case 4: len =  l; break;
   }
   if (lol) *lol=Lol;
   return len;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::copyfrom(TPGPbuffer *buffer, int length)
{
   if (length<=0)
   {
      length = buffer->length;
      buffer->rewind();
   }
   long maxlen=buffer->length-buffer->ptr;
   if (length>maxlen) length = maxlen;
   CopyFrom(buffer,length);

   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::copypacketfrom(TPGPbuffer *db, bool Rewind)
{
   if (db->ptr==db->length) return false;
   int len=db->GetPacketSize();
   if (!len) return false;
   copyfrom(db,len);
   if (Rewind) rewind();
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::EndOfKey()
{
   if (ptr==length) return false;
   byte ctb = *(char*)&data[ptr];
   return ctb==153 || ctb==149;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::copykeyfrom(TPGPbuffer *db, bool Rewind)
{
   long Ptr=db->ptr;
   if (Ptr==db->length) return false;
   TPGPbuffer *pkt = new TPGPbuffer();
   while (pkt->copypacketfrom(db,REWIND))
   {
      if (PublicSubKeyPacket(pkt->CTB))
      {
         pkt->IdText8();
         fsubalgo = pkt->algo;
         fsubbits = pkt->bits;
         pkt->rewind();
      }
      copyfrom(pkt);

      if (db->EndOfKey())
         break;

      pkt->clear();
   }
   delete pkt;
   fparsed = false;
   if (Rewind)
   {
      parsekeycert();
      rewind();
   }
   filename = db->filename;
   return db->ptr>Ptr;
}
//---------------------------------------------------------------------------
bool __fastcall TPGPbuffer::appendkeyfrom(TPGPbuffer *keys, bool Rewind)
{
   if (keys->ptr==keys->length) return false;
   Seek(0,soFromEnd);
   copykeyfrom(keys,Rewind);
   return true;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::GetIdText8()
{
   if (fidtext8.IsEmpty()) return IdText8();
   return fidtext8;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::IdText8()
{
 	 byte c;
   AnsiString text8;
   TPGPbuffer *id = new TPGPbuffer();
   keyid(id);
	 for (int i=0; i<4; i++)
   {
 	  	 c = id->data[4+i];
		 text8 += "0123456789ABCDEF" [ c >> 4  ];
		 text8 += "0123456789ABCDEF" [ c & 0xf ];
	 }
   delete id;
   return text8;
}
//---------------------------------------------------------------------------
byte __fastcall TPGPbuffer::Version()
{
   return fversion;
}
//---------------------------------------------------------------------------
int __fastcall TPGPbuffer::keyid(TPGPbuffer *KeyId)
{
  byte j;
  long PTR,Ptr=ptr;
  BUFFER *i = buf_new();
  BUFFER *k = buf_new();
  BUFFER *key = buf_new();
  BUFFER *dummy = buf_new();
  buf_append(key,data+3,length-3); buf_rewind(key);

  switch (fversion=(byte)buf_getc(key))
  {
     case 2:
     case 3:
        funixtime = buf_getl(key);    // timestamp
        fvalid = buf_geti(key);       // validity
        falgo = (byte)buf_getc(key);  // algo
        fmodbits = mpiget(key,i);     // bits
        // go ahead and get expbits and reset to proper location
        PTR = ptr; fexpbits = mpiget(key,dummy); ptr = PTR; // expbits
        break;
     case 4:
        buf_appendc(k,(char)fversion);
        funixtime = buf_getl(key);    // timestamp
        buf_appendl(k,funixtime);
        fvalid = 0;                   // validity
        falgo = (byte)buf_getc(key);  // algo
        buf_appendc(k, (char)falgo);
        //get fmodbits then set back for process below
        int PTR = ptr; fmodbits = mpiget(key,i); ptr = PTR; // bits
        if (mpinum(falgo)==0) buf_rest(k, key);
        else for (j = 0; j < mpinum(falgo); j++)
        {
           mpiget(key,i);
 	       mpiput(k,i);
        }
        buf_clear(i);
        buf_appendc(i, 0x99);
        buf_appendi(i, k->length);
        buf_cat(i, k);
        digest_sha1ex(i, i);
        break;
  }
  KeyId->clear();
  KeyId->Write(&i->data[i->length-8],8);
  buf_free(dummy);
  buf_free(key);
  buf_free(k);
  buf_free(i);
  ptr = Ptr;
  return 0;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::GetAlgoStr()
{
   AnsiString algostr = PKAlgoStr(falgo);
   if (fversion==4) algostr = PKAlgoStr(fsubalgo)+"/"+algostr;
   return algostr;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::GetBitsStr()
{
   AnsiString bitstr = s(fmodbits);
   if (fversion==4) bitstr = s(fsubbits)+"/"+bitstr;
   return bitstr;
}
//---------------------------------------------------------------------------
AnsiString __fastcall TPGPbuffer::PKAlgoStr(int algo)
{
   switch (algo)
   {
      case PGP_ES_RSA: return "RSA";
      case PGP_E_ELG : return "DH";
      case PGP_S_DSA : return "DSS";
   }
   return "UNDEFINED";
}
//---------------------------------------------------------------------------
// Misc PGP routines for keyrings/keys
//---------------------------------------------------------------------------
int __fastcall pgpencrypt(int mode, BUFFER *message, BUFFER *encr, BUFFER *sigid,
		BUFFER *pass, char *pubring, char *secring)
{
   int ret = NULL; PGPENCRYPT pgpencrypt;
   if ((pgpencrypt=(PGPENCRYPT)GetProcAddress(hMixLib,"pgp_encrypt"))!=NULL)
      ret = (pgpencrypt)(mode,message,encr,sigid,pass,pubring,secring);
   return ret;
}
//---------------------------------------------------------------------------
void __fastcall SHA1ex(byte *b, int n, byte *m)
{
   _SHA1ex sha1;
   if ((sha1=(_SHA1ex)GetProcAddress(hSSLLib,"SHA1"))!=NULL) (sha1)(b,n,m);
}
//---------------------------------------------------------------------------
AnsiString __fastcall mdalgostr(int num)
{
   AnsiString algo=s(num);
   switch (num)
   {
      case  1: algo += "MD5"; break;
      default: algo += "UNDEFINED";
   }
   return algo;
}
//---------------------------------------------------------------------------
int __fastcall pgpispacket(BUFFER *b)
{
  return ((b->data[b->ptr] >> 6) == 2 || (b->data[b->ptr] >> 6) == 3);
}
//---------------------------------------------------------------------------
int __fastcall pgpisconventional(BUFFER *b)
{
  long l;
  int type = pgppackettype(b, &l);
  buf_rewind(b);
  return (type == PGP_ENCRYPTED);
}
//---------------------------------------------------------------------------
int __fastcall pgppackettype(BUFFER *b, long *len)
{
  int ctb = buf_getc(b);
  switch (ctb >> 6)
  {
     case 2:
        /* old packet type */
        switch (ctb & 3)
        {
           case 0: *len = buf_getc(b); break;
           case 1: *len = buf_geti(b); break;
           case 2: *len = buf_getl(b); break;
           case 3: *len = b->length - b->ptr; break;
        }
        return (ctb >> 2) & 15;
     case 3:
        /* new packet type */
        *len = buf_getc(b);
        if (*len >= 192 && *len <= 223) *len = (*len - 192) * 256 + buf_getc(b) + 192; else
        if (*len == 255) *len = buf_getl(b); else
        if (*len > 223) return (-2); /* partial body length */
        return (ctb & 63);
  }
  return (-1);
}
//---------------------------------------------------------------------------
int __fastcall mpinum(int algo)
{
  switch (algo)
  {
     case PGP_ES_RSA: return (2);
     case PGP_S_DSA : return (4);
     case PGP_E_ELG : return (3);
  }
  return (0);
}
//---------------------------------------------------------------------------
int __fastcall digestmem_sha1ex(byte *b, int n, BUFFER *md)
{
  byte m[SHA_DIGEST_LENGTH];

  SHA1ex(b, n, m);
  buf_reset(md);
  buf_append(md, m, SHA_DIGEST_LENGTH);
  return (0);
}
//---------------------------------------------------------------------------
int __fastcall digest_sha1ex(BUFFER *b, BUFFER *md)
{
  return (digestmem_sha1ex(b->data, b->length, md));
}
//---------------------------------------------------------------------------
int __fastcall mpiget(BUFFER *b, BUFFER *mpi)
{
  buf_clear(mpi);
  int l=buf_geti(b);
  if (l <= 0 || b->ptr + (l + 7) / 8 > b->length) return (-1);
  buf_get(b, mpi, (l + 7) / 8);
  return (l);
}
//---------------------------------------------------------------------------
int __fastcall mpiput(BUFFER *buf, BUFFER *mpi)
{
  int l,i;
  l = mpi->length * 8;
  for (i = 7; i >= 0; i--)
    if (((mpi->data[0] >> i) & 1) == 1) {
      l -= 7 - i;
      break;
    }
  buf_appendi(buf, l);
  buf_cat(buf, mpi);
  return (l);
}
//---------------------------------------------------------------------------
int strleftex(const char *string, const char *keyword)
{
  register unsigned int i;

  for (i=0; keyword[i]!='\0'; i++)
     if (string[i] != keyword[i]) return 0;
  return 1;
}
//---------------------------------------------------------------------------
// mode=0: normal message, 1: remailer message, 2: key 3: nym key 4: nym signature
int __fastcall pgparmor(BUFFER *in, int mode)
{
#define THEPOLY 0X1864CFB
   BUFFER *out;
   unsigned long crc = 0xB704CE;
   long p;
   int i;

#if 0
   /* CRC algorithm from RFC 2440 */
   for (p = 0; p < in->length; p++)
   {
      crc ^= in->data[p] << 16;
      for (i = 0; i < 8; i++)
      {
         crc <<= 1;
         if (crc & 0x1000000) crc ^= THEPOLY;
      }
   }
#else
   /* pre-computed CRC table -- much faster */
   unsigned long table[256];
   unsigned long t;
   int q = 0;

   table[0] = 0;
   for (i = 0; i < 128; i++)
   {
      t = table[i] << 1;
      if (t & 0x1000000)
      {
         table[q++] = t ^ THEPOLY;
         table[q++] = t;
      }
      else
      {
         table[q++] = t;
         table[q++] = t ^ THEPOLY;
      }
   }
   for (p = 0; p<in->length; p++)
      crc = (crc<<8) ^ table[(in->data[p] ^ (crc>>16)) & 255];
#endif
   encode(in, 64);
   out = buf_new();
   if (mode==2 || mode==3) buf_sets(out,begin_pgpkey); else
   if (mode==4) buf_sets(out, begin_pgpsig); else
   buf_sets(out, begin_pgpmsg);
   buf_nl(out);
#if 1
   buf_appends(out, "Version: N/A\n");
#else
#ifdef MIMIC
   if (mode==1 || mode==3 || mode==4) buf_appends(out, "Version: 2.6.3i\n");
   else
#endif
   {
      buf_appends(out, "Version: Mixmaster ");
      buf_appends(out, VERSION);
      buf_appends(out, " (OpenPGP module)\n");
   }
#endif
   buf_nl(out);
   buf_cat(out, in);
   buf_reset(in);
   buf_appendc(in, (byte)((crc >> 16) & 255));
   buf_appendc(in, (byte)((crc >> 8) & 255));
   buf_appendc(in, (byte)( crc & 255));
   encode(in, 0);
   buf_appendc(out, '=');
   buf_cat(out, in);
   buf_nl(out);
   if (mode==2 || mode==3) buf_appends(out,end_pgpkey); else
   if (mode==4) buf_appends(out, end_pgpsig); else
   buf_appends(out, end_pgpmsg);

   buf_nl(out);
   buf_move(in, out);
   buf_free(out);
   return 0;
}
//---------------------------------------------------------------------------
int _fastcall __fastcall pgpdearmor(BUFFER *in, BUFFER *out)
{
  BUFFER *line, *temp;
  int tempbuf = 0;
  int err;// = 0;

  line = buf_new();
  temp = buf_new();
  if (in==out)
  {
     out = buf_new();
     tempbuf = 1;
  }
  do if (buf_getline(in, line) == -1)
  {
     err = -1;
     goto end;
  }
  while (!strleftex((char*)line->data,"-----BEGIN PGP ")) ;
  while (buf_getheader(in, temp, line) == 0) ;	/* scan for empty line */

  /*err = */decode(in, out);
  err = buf_getline(in, line);
  if (line->length == 5 && line->data[0] == '=')	/* skip CRC */
     err = buf_getline(in, line);
  if (err == 0 && line->length && strleftex((char*)line->data,"-----END PGP "))
     err = 0;
  else
     err = -1;

end:
  buf_free(temp);
  buf_free(line);
  if (tempbuf)
  {
    buf_move(in,out);
    buf_free(out);
  }
  return (err);
}
//---------------------------------------------------------------------------
int __fastcall pgpgetsize(BUFFER *packet, byte ctb)
{
   int l=0;
   if (ctb==0)
   {
      buf_rewind(packet);
      ctb = (byte)buf_getc(packet);
   }
   switch (lolf(ctb))
   {
      case 1: l = buf_getc(packet); break;
      case 2: l = buf_geti(packet); break;
      case 4: l = buf_getl(packet); break;
   }
   return l;
}
//---------------------------------------------------------------------------
AnsiString __fastcall BinaryByte(byte value, byte frombit, byte tobit)
{
   char buf[9];
   AnsiString bin=itoa(value,buf,2);
   bin = s("00000000").SubString(1,8-bin.Length())+bin;
   return bin.SubString(7-frombit+1,frombit-tobit+1);
}
//---------------------------------------------------------------------------
int __fastcall pgp_getkeyid(BUFFER *packet, BUFFER *keyid)
{
  BUFFER *i,*k;
  int version,algo,j;
  buf_rewind(packet);
  byte ctb=(byte)buf_getc(packet);
  AnsiString bin=BinaryByte(ctb,5,2);
  pgpgetsize(packet);

  if (bin=="0110" || bin=="0101") // key certificate packet
  {
     i = buf_new();
     k = buf_new();
     switch (version=buf_getc(packet))
     {
        case 2:
        case 3:
           packet->ptr += ( 4 + 2 + 1 ); // timestamp, validity, algo
           mpiget(packet,i);
           break;
        case 4:
           buf_appendc(k, (char)version);
           buf_appendl(k, buf_getl(packet));
           algo = buf_getc(packet);
           buf_appendc(k, (char)algo);
           if (mpinum(algo) == 0) buf_rest(k,packet); // works for public keys only
           else
              for (j = 0; j < mpinum(algo); j++)
              {
 	               mpiget(packet,i);
 	               mpiput(k,i);
              }
           buf_clear(i);
           buf_appendc(i, 0x99);
           buf_appendi(i, k->length);
           buf_cat(i, k);
           digest_sha1ex(i, i);
           break;
     }
     buf_clear(keyid);
     buf_append(keyid, i->data + i->length - 8, 8);
     buf_free(i);
     buf_free(k);
     return 0;
  }
  return 0;
}
//---------------------------------------------------------------------------
AnsiString __fastcall pgp_getkeyidasc(BUFFER *packet)
{
   byte c;
   AnsiString text8;
   BUFFER *id = buf_new();
   pgp_getkeyid(packet,id);
   for (int i=0; i<4; i++)
   {
   	  c = id->data[4+i];
 	     text8 += "0123456789ABCDEF" [ c >> 4  ];
	     text8 += "0123456789ABCDEF" [ c & 0xf ];
   }
   buf_free(id);
   return text8;
}
//---------------------------------------------------------------------------
/* Format of public key certificate
*
*   Ofs  Len  Meaning
*   0    1    CTB for public key certificate
*   1    2    16-bit (or maybe 8-bit) length of packet
*   3	 1	  Version byte, may affect rest of fields that follow.
*                 (=2) for PGP versions <= 2.5
*                 (=3) for PGP versions >= 2.6
*   4    4    Timestamp of key creation
*   8    2    Validity period, in number of DAYS (0 means forever)
*   10	 1	  Algorithm byte for RSA (=1 for RSA).
*    		  	--Algorithm byte affects field definitions that follow.
*   ?    ?    MPI of RSA public modulus n
*   ?    ?    MPI of RSA public encryption exponent e
*
*/
bool __fastcall ParsePublicKeyPacket(BUFFER *packet, AnsiString *idtext8, int *version,
   int *unixtime, int *valid, int *algo, int *modbits, int *expbits,
   BUFFER *modulus, BUFFER *exponent)
{
   if (idtext8) *idtext8 = pgp_getkeyidasc(packet);
   buf_rewind(packet);
   byte ctb=(byte)buf_getc(packet);
   if (BinaryByte(ctb,5,2)!="0110") return false;
   pgpgetsize(packet);

   long ver,*ptr = &packet->ptr;
   ver = buf_getc(packet); if (version) *version = ver;
   if (unixtime) *unixtime = buf_getl(packet); else *ptr+=4;
   switch (ver)
   {
      case 2:
      case 3: if (valid) *valid = buf_geti(packet); else *ptr+=2; break;
      case 4: if (valid) *valid = buf_getc(packet); else *ptr+=1; break;
      default: -1;
   }
   if (algo) *algo = buf_getc(packet); else *ptr+=1;

   long Bits;
   BUFFER *dummy = buf_new();
   if (modulus) Bits = mpiget(packet,modulus);
   else Bits = mpiget(packet,dummy);
   if (modbits) *modbits = Bits;

   if (exponent) Bits = mpiget(packet,exponent);
   else Bits = mpiget(packet,dummy);
   if (expbits) *expbits = Bits;
   buf_free(dummy);

   return true;
}
//---------------------------------------------------------------------------
/* Format of secret key certificate
*
*   Offset  Length  Meaning
*   0       1       CTB for secret key certificate
*   1       2       16-bit (or maybe 8-bit) length of packet
*   3	    1	      Version byte, may affect rest of fields that follow.
*                   (=2) for PGP versions <= 2.5
*                   (=3) for PGP versions >= 2.6
*   4       4       Timestamp
*   8       2       Validity period, in number of DAYS (0 means forever)
*  10	    1	      Algorithm byte for RSA (=1 for RSA).
*   		          --Algorithm byte affects field definitions that follow.
*   ?       ?       MPI of RSA public modulus n
*   ?       ?       MPI of RSA public encryption exponent e
*   ?	    1	      Algorithm byte for cipher that protects following
*   		            secret components (0=unencrypted, 1=IDEA cipher)
*   ?	    8	      Cipher Feedback IV for cipher that protects secret
*          		   components (not present if unencrypted)
*   ?       ?       MPI of RSA secret decryption exponent d
*   ?       ?       MPI of RSA secret factor p
*   ?       ?       MPI of RSA secret factor q
*   ?       ?       MPI of RSA secret multiplicative inverse u
*                  (All MPI's have bitcount prefixes)
*   ?	    2	      16-bit checksum of all preceding secret component bytes
*
*/
bool __fastcall ParseSecretKeyPacket(BUFFER *packet, AnsiString *idtext8, byte *version,
   int *unixtime, int *valid, byte *algo, int *modbits, int *expbits, BUFFER *modulus,
   BUFFER *exponent, int *algocip, BUFFER *cipfbIV, BUFFER *dexponent, int *dexbits,
   BUFFER *factorp, int *facpbits, BUFFER *factorq, int *facqbits,
   BUFFER *inverse, int *invbits, int *checksum)
{
   AnsiString id = pgp_getkeyidasc(packet);
   if (idtext8) *idtext8 = pgp_getkeyidasc(packet);

   buf_rewind(packet);
   byte ctb=(byte)buf_getc(packet);
   if (BinaryByte(ctb,5,2)!="0101") return false;
   pgpgetsize(packet);

   int Bits;
   byte alg,ver;
   BUFFER *dummy = buf_new();
   ver = (byte)buf_getc(packet); if (version) *version = ver;
   if (unixtime) *unixtime = buf_getl(packet); else packet->ptr+=4;
   switch (ver)
   {
      case 2:
      case 3: if (valid) *valid = buf_geti(packet); else packet->ptr+=2; break;
      case 4: if (valid) *valid = buf_getc(packet); else packet->ptr+=1; break;
      default: -1;
   }
   alg = (byte)buf_getc(packet); if (algo) *algo = alg;

   if (modulus) Bits = mpiget(packet,modulus);
   else Bits = mpiget(packet,dummy);
   if (modbits) *modbits = Bits;

   if (exponent) Bits = mpiget(packet,exponent);
   else Bits = mpiget(packet,dummy);
   if (expbits) *expbits = Bits;

   alg = (byte)buf_getc(packet); if (algocip) *algocip = alg;

   if (cipfbIV) buf_get(packet,cipfbIV,8);
   else buf_get(packet,dummy,8);

   if (!alg) return true;

   if (dexponent) Bits = mpiget(packet,dexponent);
   else Bits = mpiget(packet,dummy);
   if (dexbits) *dexbits = Bits;

   if (factorp) Bits = mpiget(packet,factorp);
   else Bits = mpiget(packet,dummy);
   if (facpbits) *facpbits = Bits;

   if (factorq) Bits = mpiget(packet,factorq);
   else Bits = mpiget(packet,dummy);
   if (facqbits) *facqbits = Bits;

   if (inverse) Bits = mpiget(packet,inverse);
   else Bits = mpiget(packet,dummy);
   if (invbits) *invbits = Bits;

   int chksum = buf_getl(packet); if (checksum) *checksum = chksum;

   buf_free(dummy);

   return true;
}
//---------------------------------------------------------------------------
/* Format of regular signature packets: v2, v3
*
*	Offset Length Meaning
*	0 1 Version byte (= 2 or 3).
*	1 1 x, Length of following material included in MD5 (=5)
*	2 1 Signature type (=0 or 1)
*	3 4 32-bit timestamp of signature
* -------- MD5 additional material stops here, at offset 2+x ---------
*	2+x	8	KeyID
*	10+x	1	PK algorithm type (1 = RSA)
*	11+x	1	MD algorithm type (1 = MD5)
*	12+x	2	First 2 bytes of message digest (16-bit checksum)
*	14+x	2+y	MPI of PK-encrypted integer
*	16+x+y
*/
bool __fastcall ParseSignaturePacket3(BUFFER *packet, int *version,
   int *implicit, int *sigclass, int *unixtime,
   AnsiString *idtext8, int *pkalgo,
   int *mdalgo, int *chksum)
{
   buf_rewind(packet);
   byte ctb=(byte)buf_getc(packet);
   if (BinaryByte(ctb,5,2)!="0010") return false;
   pgpgetsize(packet);

   if (version) *version = buf_getc(packet); else packet->ptr += 1;
   if (implicit) *implicit = buf_getc(packet); else packet->ptr += 1;
   if (sigclass) *sigclass = buf_getc(packet); else packet->ptr += 1;
   if (unixtime) *unixtime = buf_getl(packet); else packet->ptr += 4;
   if (idtext8) *idtext8 = pgp_getkeyidasc(packet); else packet->ptr += 8;
   if (pkalgo) *pkalgo = buf_getc(packet); else packet->ptr += 1;
   if (mdalgo) *mdalgo = buf_getc(packet); else packet->ptr += 1;
   if (chksum) *chksum = buf_geti(packet); else packet->ptr += 2;

   return true;
}
//---------------------------------------------------------------------------
/* Format of one-pass signature packets:  v4
*
*	Offset Length Meaning
*	0	1	Version byte (=3)
*	1	1	Signature type
*	2	1	Hash Algorithm
*	3	1	PK Algorithm
*	4	8	KeyID
*	12	1	nested flag
*/
bool __fastcall ParseSignaturePacket4(BUFFER *packet, int *version,
   int *sigclass, int *hashalgo, int *pkalgo,
   AnsiString *idtext8, int *nestflag)
{
   buf_rewind(packet);
   byte ctb=(byte)buf_getc(packet);
   if (BinaryByte(ctb,5,2)!="0010") return false;
   pgpgetsize(packet);

   if (version) *version = buf_getc(packet); else packet->ptr += 1;
   if (sigclass) *sigclass = buf_getc(packet); else packet->ptr += 1;
   if (hashalgo) *hashalgo = buf_getc(packet); else packet->ptr += 1;
   if (pkalgo) *pkalgo = buf_getc(packet); else packet->ptr += 1;
   if (idtext8) *idtext8 = pgp_getkeyidasc(packet); else packet->ptr += 8;
   if (nestflag) *nestflag = buf_getc(packet); else packet->ptr += 1;

   return true;
}
//---------------------------------------------------------------------------
bool __fastcall ParseUserIdPacket(BUFFER *packet, AnsiString &userid)
{
   userid = (char*)&packet->data[2];
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall ParseCommentPacket(BUFFER *packet, AnsiString &comment)
{
   //BUFFER *buf=buf_new();
   //int len=buf_getc(packet);
   //buf_get(packet,buf,len);
   //comment = (char*)&buf->data[0];
   //buf_free(buf);
   comment = (char*)&packet->data[2];
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall ParseTrustPacket(BUFFER *packet, AnsiString &trust)
{
   //byte ctb=buf_getc(packet);
   //sm((int)pgp_getsizeex(packet,ctb));
   char binstr[17];
   int i = packet->data[2];
   AnsiString trs = itoa(i,binstr,2);
   trust = s("00000000").SubString(1,8-trs.Length())+trs;
   return true;
}
//---------------------------------------------------------------------------
bool __fastcall ParseLiteralDataPacket(BUFFER *packet,
   int &mode, AnsiString &filename, int &unixtime)
{
   buf_rewind(packet);
   byte ctb=(byte)buf_getc(packet);
   if (BinaryByte(ctb,5,2)!="1011") return false;

   pgpgetsize(packet);
   mode = buf_getc(packet);
   int len = buf_getc(packet);

   BUFFER *buf=buf_new();
   buf_get(packet,buf,len);
   filename = (char*)buf->data;
   delete buf;

   unixtime = buf_getl(packet);
   return true;
}
//---------------------------------------------------------------------------
